--| DICTIONARY
--|-----------------------------------------------------------------------------
--|   Copyright (C) GunnarGotshalks --  All rights reserved.
--|
--| Permission to use, copy, modify, and distribute,without fee this
--| software and its documentation for educational purposes and is
--| hereby granted.
--|-----------------------------------------------------------------------------

note
  description: "Dictionary for <key, value> pairs"
  author: "Gunnar Gotshalks"
  date: "2003 February 10"
  version: "1.0"

class DICTIONARY [KEY -> HASHABLE  , VALUE]

create make

feature -- Creation

make (theCapacity : INTEGER)
  -- Create an empty Dictionary of size theCapacity
  require
  	minimum_capacity: theCapacity >= minimum_capacity
  local
  	initial_key : KEY        -- At generation time correct initial values are
  	initial_value : VALUE    -- used in the following create routines.
  do
    create keys.make_filled (initial_key, 1, theCapacity)
    create values.make_filled (initial_value, 1, theCapacity)
    create used.make_filled (false, 1, theCapacity)
    capacity := theCapacity
    count := 0
  ensure
  	capacity_set: capacity = theCapacity
  	zero_count: count = 0
  	representation_exists: keys /= void and values /= void and used /= void
  end

--------------------------------------------------------------------------------
feature {NONE} -- Representation data
    -- A <key, value> pair is stored in corresponding array positions in
    -- keys and values respectively.

keys : ARRAY[KEY]
values : ARRAY[VALUE]
used : ARRAY[BOOLEAN]  -- True indicates the corresponding <key, value> exists.

--------------------------------------------------------------------------------
feature -- Meta data

minimum_capacity : INTEGER  once Result := 10 end
                     -- Provide at least this much space in the dictionary
capacity: INTEGER    -- Maximum number of pairs that can be held
count: INTEGER       -- Number of pairs in the dictionary

--------------------------------------------------------------------------------
feature -- Query operations

has (key : KEY) : BOOLEAN
  -- True if a <key, don't_care> pair is in the dictionary
  require
  	key_exists: key /= void
  do
   	Result := location_of(key) /= 0
  ensure
    has_ok: Result = (location_of(key) /= 0)
    no_change: no_change_attributes(old capacity, old count,
                 old keys.deep_twin, old values.deep_twin,
                 old used.deep_twin)
  end

at alias "@" (key : KEY) : VALUE
  -- Return the value in the <key , value> pair
  require
  	key_exists: key /= void
  	in_dictionary: has(key)
  do
    Result := values @ location_of(key)
  ensure
    at_key_ok: Result = values @ location_of(key)
         -- Depends upon representation.  Expression changes as representation
         -- changes.
    no_change: no_change_attributes(old capacity, old count,
                 old keys.deep_twin, old values.deep_twin,
                 old used.deep_twin)
  end

is_empty : BOOLEAN
  -- True if and only if the dictionary is empty
  require True
  do Result := count = 0
  ensure
    is_empty: Result = (count = 0)
    no_change: no_change_attributes(old capacity, old count,
                 old keys.deep_twin, old values.deep_twin,
                 old used.deep_twin)
  end

is_full : BOOLEAN
  -- True if and only if the dictionary is full
  require True
  do Result := count = capacity
  ensure
    is_full: Result = (count = capacity)
    no_change: no_change_attributes(old capacity, old count,
                 old keys.deep_twin, old values.deep_twin,
                 old used.deep_twin)
  end

--------------------------------------------------------------------------------
feature -- Write operations

put (key : KEY; value : VALUE)
  -- Insert the <key,value> pair into the dictionary only if an entry with the
  -- key does not already exist in the dictionary
  require
  	key_exists: key /= void
  	not_in_dictionary: not has(key)
  	space_exists: count < capacity
  local k : INTEGER -- index to keys and values
  do
  	k := empty_entry
  	keys.put(key,k) ; values.put(value,k) ; used.put(true,k)
  	count := count + 1
  ensure
  	put_ok: at(key) = value
  	count_incremented: count = old count + 1
  end

put_defensive (key : KEY; value : VALUE)
  -- If the key does not exist in the dictionary, then add the <key,value>
  -- entry.  Otherwise, replace the existing value for the key.
  require
  	key_exists: key /= void
  	space_exists_defensive: has(key) or count < capacity
  do
  	if has(key) then replace(key, value)
  	else put(key,value)
  	end
  ensure
    put_defensive_ok: at(key) = value
    potential_count_increase: ( old current.has(key) and count = old count) xor
                              (not old current.has(key) and count = old count + 1)
  end

replace (key : KEY; value : VALUE)
  -- Insert the <key,value> pair into the dictionary only if an entry with the
  -- key does not already exist in the dictionary
  require
  	key_exists: key /= void
  	in_dictionary: has(key)
  do
  	values.put(value, location_of(key))
  ensure
  	replace_ok: at(key) = value
  	count_unchanged: count = old count
  end

remove (key : KEY)
  -- Remove the <key,???> pair from the dictionary
  require
  	key_exists: key /= void
  	in_dictionary: has(key)
  do
  	used.put(false, location_of(key))
  	count := count - 1
  ensure
  	no_entry: not has(key)
  	decrease_count: count = old count - 1
  end

--------------------------------------------------------------------------------
feature {NONE} -- Representation operations
--------------------------------------------------------------------------------

location_of(key : KEY) : INTEGER
  -- Return the location for a key.  representation dependent but do
  -- expect 0 if key not in representation.s
  require key /= void
  local k : INTEGER -- index to keys and values
  do
    from k := 1
    invariant -- for j : 1 .. k-1 :: used @ j --> keys @ j /= key
    until k > capacity or
         (used @ k and  key.is_equal(keys @ k))
    loop k := k + 1
    variant capacity + 1 - k
    end
  if k > capacity then Result := 0
  else Result := k
  end
  ensure
  	location_of_ok: Result = 0 -- only if key is not in representation
  	             or key.is_equal(keys @ Result)
  end

empty_entry : INTEGER
  -- Returns an index of an unused dictionary entry location
  require
  	space_exists: count < capacity
  do
    from Result := 1
    invariant -- forall j : 1..Result-1 :: used(keys @ j)
    until not (used @ Result)
    loop Result := Result + 1
    end
  ensure
  	point_to_empty: not(used @ Result)
  end

no_change_attributes( old_capacity, old_count : INTEGER
                    ; old_keys : ARRAY[KEY]; old_values : ARRAY[VALUE]
                    ; old_used : ARRAY[BOOLEAN]) : BOOLEAN
 do
 	Result := capacity = old_capacity
 	      and count = old_count and deep_equal(keys, old_keys)
 	      and deep_equal(values, old_values)
 	      and deep_equal(used, old_used)
 end

invariant
  count_not_negative: count >= 0
  count_no_greater_than_capacity: count <= capacity
  not_empty: count > 0 implies not is_empty
  not_full: count < capacity implies not is_full

end
